# Perl limited interface for the Azure AD Graph API
#
# This Perl module provides a Perl interface to the Azure AD Graph API.
# It's meant as a very low-coverage module, since there is currently no
# standard module and our own needs are fairly small.

package Stanford::Azure::Graph 1.00;

use 5.014;
use strict;
use warnings;

use Carp qw(croak);
use HTTP::Request;
use JSON ();
use LWP::UserAgent 6.00;
use Stanford::Azure::Graph::Token;
use Stanford::Azure::Graph::Exception;
use Perl6::Slurp;
use POSIX qw(strftime);
use URI::Escape qw(uri_escape_utf8);

my $API_VERSION = '1.5';

# All dies are of constructed objects, which perlcritic misdiagnoses.
## no critic (ErrorHandling::RequireCarping)

##############################################################################
# Constructor
##############################################################################

# Create a new Stanford::Azure::Graph object, which will be used for
# subsequent calls.
#
# $class - Class of object to create
# $args  - Anonymous hash of arguments with the following keys:
#   api_hostname    - API hostname for the Graph API integration
#   integration_key - Public key for the Graph API integration
#   key_file        - Path to file with integration information
#   secret_key      - Secret key for the Graph API integration
#   user_agent      - User agent object to use instead of LWP::UserAgent
#
# Returns: Newly-created object
#  Throws: Stanford::Azure::Graph::Exception on any failure
sub new {
    my ($class, $args_ref) = @_;
    my $self = {};

    # Load integration information from key_file if set.
    my $keys;
    if ($args_ref->{key_file}) {
        my $json     = JSON->new->relaxed(1);
        my $key_data = slurp($args_ref->{key_file});
        $keys = eval { $json->decode($key_data) };
        if ($@) {
            die Stanford::Azure::Graph::Exception->propagate($@);
        }
    }

    # Integration data from $args_ref overrides key_file data.
    for my $key (qw(client_id client_secret scope)) {
        $self->{$key} = $args_ref->{$key} // $keys->{$key};
        if (!defined($self->{$key})) {
            my $error = "missing parameter to Stanford::Azure::Graph::new: $key";
            die Stanford::Azure::Graph::Exception->internal($error);
        }
    }

    # Get the token provider.  Use this rather than straight token since it
    # can refresh the token as needed.
    my $token_provider
        = Stanford::Azure::Graph::Token->new($self->{client_id},
                                             $self->{client_secret},
                                             $self->{scope});
    $self->{token_provider} = $token_provider;

    # Create or set the user agent object.
    $self->{agent} = $args_ref->{user_agent} // LWP::UserAgent->new;

    # Create the JSON decoder that we'll use for subsequent operations.
    $self->{json} = JSON->new->utf8(1);

    # Bless and return the new object.
    bless($self, $class);
    return $self;
}

##############################################################################
# General methods
##############################################################################

# Make a generic Azure API call with no assumptions about its return data type.
# This returns the raw HTTP::Response object without any further processing.
# For most Azure APIs, use call_json instead, which assumes that the call
# returns JSON in a particular structure and checks the status of the HTTP
# response.
#
# $self     - Stanford::Azure::Graph object
# $method   - HTTP method (GET, POST, PATCH, or DELETE)
# $path     - URL path to the REST endpoint to call
# $args_ref - Reference to a hash of additional arguments
#
# Returns: The HTTP::Response object from the API call
#  Throws: Stanford::Azure::Graph::Exception on any failure
sub call {
    my ($self, $method, $path, $args_ref) = @_;
    my $host   = 'graph.windows.net';
    my $domain = $self->{scope};
    my $token  = $self->{token_provider}->token;
    $method    = uc($method);

    # Change any arguments given into a JSON string for posting.
    my $args;
    if ($args_ref) {
        $args = $self->{json}->encode($args_ref);
    }

    # Verify that the URL path starts with a slash.
    if ($path !~ m{ \A / }xms) {
        my $error = "REST endpoint '$path' does not begin with /";
        die Stanford::Azure::Graph::Exception->internal($error);
    }

    # Set up the request.
    my $request = HTTP::Request->new;
    $request->method($method);
    $request->protocol('HTTP/1.1');
    $request->date(time());
    $request->header(Host => $host);
    $request->authorization('Bearer ' . $token);
    $request->content_type('application/json');

    # Build the URI.  GET should never have any arguments based on how they
    # do the API.
    my $uri = 'https://' . $host . '/' . $domain . $path . '?api-version='
        . $API_VERSION;
    if ($method eq 'POST' || $method eq 'PATCH' || $method eq 'PUT') {
        $request->content($args);
    }
    $request->uri($uri);

    # Make the request and return the response.
    return $self->{agent}->request($request);
}

# Make a generic Azure API call that returns JSON and do the return status
# checking that's common to most of the Azure API calls.  There are a few
# exceptions, like /logo, which do not return JSON and therefore cannot be
# called using this method).
#
# $self     - Stanford::Azure::Graph object
# $method   - HTTP method (GET, PUT, POST, or DELETE)
# $path     - URL path to the REST endpoint to call
# $args_ref - Reference to a hash of additional arguments
#
# Returns: Reference to hash corresponding to the JSON result
#  Throws: Stanford::Azure::Graph::Exception on any failure
sub call_json {
    my ($self, $method, $path, $args_ref) = @_;

    # Use the simpler call() method to do most of the work.  This returns the
    # HTTP::Response object.  Retrieve the content of the response as well.
    my $response = $self->call($method, $path, $args_ref);
    my $content = $response->decoded_content;

    # If the response failed, treat as an HTTP failure.
    if (!$response->is_success) {
        #die Stanford::Azure::Graph::Exception->http($response);
        print $content, "\n";
    }

    # Otherwise, try to decode the JSON.  If we cannot, treat this as an
    # HTTP failure if we didn't get success and a protocol failure
    # otherwise.
    my $data = {};
    if ($content) {
        $data = eval { $self->{json}->decode($content) };
        if ($@) {
            if ($response->is_success) {
                my $error = 'invalid JSON in reply';
                die Stanford::Azure::Graph::Exception->protocol($error, $content);
            } else {
                die Stanford::Azure::Graph::Exception->http($response);
            }
        }
    }

    # Return the decoded JSON.
    return $data;
}

1;
__END__

=head1 NAME

Stanford::Azure::Graph - API for Azure Graph Active Directory service

=head1 SYNOPSIS

    my $api = Stanford::Azure::Graph->new({ key_file => '/path/to/keys.json' });
    my $reply = $api->call_json('GET', '/users');

=head1 REQUIREMENTS

Perl 5.14 or later and the modules HTTP::Request and HTTP::Response (part
of HTTP::Message), JSON, LWP (also known as libwww-perl), Perl6::Slurp,
and URI::Escape (part of URI), all of which are available from CPAN.

=head1 DESCRIPTION

Stanford::Azure::Graph provides an object-oriented Perl interface for generic
calls to the Azure AD Graph REST APIs.  This module is intended primarily
for use as a base class for more specialized Perl modules implementing the
specific Graph APIs, but it can also be used directly to make generic API
calls.

Nothing in this module is specific to Stanford, however we have no need or
intent to make a general interface for the Graph REST API and so are limiting
the namespace.  If someone else works on a more general API module, then we
will likely move to it instead.

On failure, all methods throw a Stanford::Azure::Graph::Exception object.
This can be interpolated into a string for a simple error message, or
inspected with method calls for more details.  This is also true of all
methods in all objects in the Stanford::Azure::Graph namespace.

=head1 CLASS METHODS

=over 4

=item new(ARGS)

Create a new Stanford::Azure::Graph object.  This should be used for all
subsequent API calls.  ARGS should be a hash reference with one or more of
the following keys:

=over 4

=item api_hostname

The API hostname returned by Azure when the API integration was created.

This key is required if C<key_file> is not set.

=item integration_key

The integration key returned by Azure when the API integration was created.
This is effectively the public "username" for this integration.

This key is required if C<key_file> is not set.

=item key_file

The path to a file in JSON format that contains the key and hostname data
for a Azure integration.  This file should contain one JSON object with keys
C<integration_key>, C<secret_key>, and C<api_hostname>.  These are the
three data values that are returned when one creates a new Azure API
integration.

Be aware that the C<secret_key> value in this file is security-sensitive
information equivalent to a password.  Anyone in possession of that key
has complete control of all data and actions to which the integration has
access.

Either this key or all of C<integration_key>, C<secret_key>, and
C<api_hostname> must be provided.  If both this key and some of those
keys are provided, their values will override the values retrieved from
the C<key_file> file.

=item secret_key

The secret key returned by Azure when the API integration was created.  This
is security-sensitive information equivalent to a password.  Anyone in
possession of that key has complete control of all data and actions to
which the integration has access.  Do not hard-code this into programs;
instead, read it from a file with appropriate permissions or retrieve it
via some other secure mechanism.

This key is required if C<key_file> is not set.

=item user_agent

The user agent to use for all requests.  This should be a Perl object that
supports the same API as LWP::UserAgent.

Normally, the caller will not provide this key, in which case Net::Azure will
create an LWP::UserAgent object internally to use to make Azure API calls.
This argument is provided primarily so that the user agent can be overridden
for unit testing.

=back

=back

=head1 INSTANCE METHODS

=over 4

=item call(METHOD, PATH[, ARGS])

Make a generic Graph API call, with no assumptions about the response.

This is a low-level escape hatch to make any Graph API call that this module
does not know about, regardless of what format in which it returns its
results.  The caller will have to provide all of the details (HTTP method,
URL path, and all arguments as a reference to a hash of key/value pairs).
The URL path must start with a slash.

The return value is the resulting HTTP::Response object from the web API
call.  No error checking will be performed.  The caller is responsible for
examining the HTTP::Response object for any problems, including internal
or HTTP errors.

Most Graph API calls return structured JSON.  For those calls, use
call_json() instead of this method.  call() is needed only for the small
handful of API calls that do not return JSON in that format.

=item call_json(METHOD, PATH[, ARGS])

Make a generic Graph API call that returns a JSON response.

This is the escape hatch to use to make any Graph API call that this module
does not know about.  The caller will have to provide all of the details
(HTTP method, URL path, and all arguments as a reference to a hash of
key/value pairs).  The URL path must start with a slash.

The return value will be the hashref of the returned JSON.

=back

=head1 AUTHOR

Jon Robertson <jonrober@stanford.edu>

=head1 COPYRIGHT AND LICENSE

Copyright 2015 The Board of Trustees of the Leland Stanford Junior
University

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.

=head1 SEE ALSO

L<Azure AD Graph REST API|https://msdn.microsoft.com/en-us/library/azure/hh974478.aspx>

=cut
