# Perl interface for the AD Graph User API operations
#
# This Perl module collection provides a Perl interface to the Graph API
# integration to the Azure AD system.

package Stanford::Azure::Graph::User 1.00;

use 5.014;
use strict;
use warnings;

use parent qw(Stanford::Azure::Graph);

use Carp qw(croak);
use JSON ();
use Stanford::Azure::Graph::Exception;
use POSIX qw(strftime);

##############################################################################
# Misc Methods
##############################################################################

sub _password {
    my $passwd = `/usr/bin/apg -n 1 -M SNCL -m 20 -x 25`;
    chomp $passwd;
    return $passwd;
}

##############################################################################
# API Methods
##############################################################################

# Create a new User object.  Calls the base class in order to load needed
# functions as well as the authtoken.
sub new {
    my ($class, $data_ref) = @_;
    return $class->SUPER::new($data_ref);
}

# Create a new user in AD with minimal given information.
#
# $principal   - Principal to create (email address)
# $displayname - Name to display for user
# $nickname    - ????
#
# Returns - hashref of user data on success
#           undef and warning on failure
sub create {
    my ($self, $principal, $displayname, $nickname, $immutableid,
        $license) = @_;

    my %license_rec = (skuId         => $license,
                       disabledPlans => [],
                      );

    my $enabled = JSON::true;
    my %args = (accountEnabled    => $enabled,
                displayName       => $displayname,
                mailNickname      => $nickname,
                userPrincipalName => $principal,

                passwordProfile   => {
                    password                     => _password(),
                    forceChangePasswordNextLogin => JSON::false,
                },

                usageLocation     => 'US',

                immutableId       => $immutableid,
               );

    # Do we create enabled or disabled and let things pick up?
    # Assign licenses right now?

    my $data;
    eval { $data = $self->call_json('POST', '/users', \%args) };
    if ($@) {
        warn "could not create user: $@\n";
        return;
    }

    return $data;
}

# Get all user data in AD for a principal.
#
# $principal - Principal to create (email address)
#
# Returns - hashref of user data on success
#           undef and warning on failure
sub user {
    my ($self, $principal) = @_;

    my $data;
    eval { $data = $self->call_json('GET', '/users/'.$principal) };
    if ($@) {
        die "could not get user: $@";
    }

    # Check for errors and return undef in either case.  If it's any error
    # other than the user not being found, print a warning.
    if ($data->{'odata.error'}) {
        if ($data->{'odata.error'}{code} ne 'Request_ResourceNotFound') {
            die $data->{'odata.error'}{code};
        } else {
            return;
        }
    }

    return $data;
}

# Get all user data in AD for a principal.
#
# $principal - Principal to create (email address)
#
# Returns - hashref of user data on success
#           undef and warning on failure
sub users {
    my ($self) = @_;

    my $data;
    eval { $data = $self->call_json('GET', '/users') };
    if ($@) {
        warn "could not get user: $@\n";
        return;
    }

    return $data;
}

# Updates information for a principal in AD.
#
# $principal - Principal to delete (email address)
# $updates   - Hashref of any fields and data to update
#
# Returns - 1 on success
#           0 and warning on failure
sub update {
    my ($self, $principal, $updates) = @_;

    my $data;
    eval { $data = $self->call_json('PATCH', '/users/'.$principal, $updates) };
    if ($@) {
        warn "could not update user: $@\n";
        return 0;
    }

    return 1;
}

# Purge an existing user principal from AD
#
# $principal - Principal to delete (email address)
#
# Returns - 1 on success
#           0 and warning on failure
sub purge {
    my ($self, $principal) = @_;

    my $data;
    eval { $data = $self->call_json('DELETE', '/users/'.$principal) };
    if ($@) {
        warn "could not purge user: $@\n";
        return 0;
    }

    return 1;
}

# Assign one or more licenses to a user.
#
# $principal - Principal to add licenses to (email address)
#
# Returns - 1 on success
#           0 and warning on failure
sub assign_licenses {
    my ($self, $principal, $assigned, $removed, $disabled_plans) = @_;

    my %request = (addLicenses    => [],
                   removeLicenses => $removed,
                  );
    for my $license (@{ $assigned }) {
        my %rec = (skuId         => $license,
                   disabledPlans => $disabled_plans,
                  );
        push (@{ $request{addLicenses} }, \%rec);
    }

    my $data;
    my $url = '/users/'.$principal.'/assignLicense';
    eval { $data = $self->call_json('POST', $url, \%request) };
    if ($@) {
        warn "could not assign license to user: $@\n";
        return 0;
    }

    return 1;
}

1;
